//
//  main.c
//  z-wave-test
//
//  Created by Alex Skalozub on 1/6/12.
//  Copyright (c) 2012 Z-Wave.Me
//

#include <stdio.h>
#include <ZWayLib.h>
#include <ZLogging.h>
#include <errno.h>

int do_work(ZWay zway);

void test_memory(ZWay zway)
{
    ZWError r;
    
    ZWBYTE buff1[] = { 0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 13, 14, 15 };
    
    r = zway_fc_memory_put_byte(zway, 0x1234, 0x42, NULL, NULL, NULL);
    r = zway_fc_memory_get_byte(zway, 0x1234, NULL, NULL, NULL);
    r = zway_fc_memory_put_buffer(zway, 0x2345, 16, buff1, NULL, NULL, NULL);
    r = zway_fc_memory_get_buffer(zway, 0x2345, 16, NULL, NULL, NULL);
}

void test_save(ZWay zway)
{
    ZWError r;
    
    ZWBYTE *data = NULL;
    size_t length = 0;
    r = zway_controller_config_save(zway, &data, &length);
    if (r == NoError)
    {
        printf("Successfully saved! (size = %lu)\n", (unsigned long int) length);
        
        char path[MAX_PATH];
        strcpy(path, "/tmp/config.tgz");
        
        FILE *t = fopen(path, "wb");
        if (t != NULL)
        {
            fwrite(data, 1, length, t);
            fclose(t);
            
            printf("Config written to %s\n", path);
        }
        else
        {
            printf("Config write error: %s\n", strerror(errno));
        }
        
        free(data);
    }
    else
    {
        printf("Error saving configuration! (err = %s)\n", zstrerror(r));
    }
}

void test_restore(ZWay zway)
{
    char path[MAX_PATH];
    strcpy(path, "/tmp/config.tgz");
    
    printf("Reading config from %s\n", path);
    
    struct stat finfo;
    int tr = stat(path, &finfo);
    if (tr != 0)
    {
        if (errno == ENOENT)
        {
            printf("Config not found. Save one first!\n");
        }
        else
        {
            printf("Config read error: %s\n", strerror(errno));
        }
        return;
    }
    
    size_t length = finfo.st_size;
    ZWBYTE *data = (ZWBYTE*)malloc(length);
    if (data == NULL)
    {
        printf("Failed to alloc %zu bytes!\n", length);
        return;
    }
    
    FILE *t = fopen(path, "rb");
    if (t != NULL)
    {
        fread(data, 1, length, t);
        fclose(t);
    }
    else
    {
        free(data);
        printf("Config read error: %s\n", strerror(errno));
        return;
    }
    
    ZWError r = zway_controller_config_restore(zway, data, length, FALSE);
    if (r == NoError)
    {
        printf("Configuration successfully restored\n");
    }
    else
    {
        printf("Error restoring configuration! (err = %s)\n", zstrerror(r));
    }
    
    free(data);
}

void dump_data(const ZWay zway, ZDataHolder data)
{
    char *path = zdata_get_path(data);
    ZWDataType type;
    zdata_get_type(data, &type);

    ZWBOOL bool_val;
    int int_val;
    float float_val;
    ZWCSTR str_val;
    const ZWBYTE *binary;
    const int *int_arr;
    const float *float_arr;
    const ZWCSTR *str_arr;
    size_t len, i;

    switch (type) 
    {
        case Empty:
            zway_log(zway, Debug, ZSTR("DATA %s = Empty"), path);
            break;
        case Boolean:
            zdata_get_boolean(data, &bool_val);
            if (bool_val)
                zway_log(zway, Debug, ZSTR("DATA %s = True"), path);
            else
                zway_log(zway, Debug, ZSTR("DATA %s = False"), path);
            break;
        case Integer:
            zdata_get_integer(data, &int_val);
            zway_log(zway, Debug, ZSTR("DATA %s = %d (0x%08x)"), path, int_val, int_val);
            break;
        case Float:
            zdata_get_float(data, &float_val);
            zway_log(zway, Debug, ZSTR("DATA %s = %f"), path, float_val);
            break;
        case String:
            zdata_get_string(data, &str_val);
            zway_log(zway, Debug, ZSTR("DATA %s = \"%s\""), path, str_val);
            break;
        case Binary:
            zdata_get_binary(data, &binary, &len);
            zway_log(zway, Debug, ZSTR("DATA %s = byte[%d]"), path, (int)len);
            zway_dump(zway, Debug, ZSTR("  "), len, binary);
            break;
        case ArrayOfInteger:
            zdata_get_integer_array(data, &int_arr, &len);
            zway_log(zway, Debug, ZSTR("DATA %s = int[%d]"), path, (int)len);
            for (i = 0; i < len; i++)
                zway_log(zway, Debug, ZSTR("  [%02d] %d"), (int)i, int_arr[i]);
            break;
        case ArrayOfFloat:
            zdata_get_float_array(data, &float_arr, &len);
            zway_log(zway, Debug, ZSTR("DATA %s = float[%d]"), path, (int)len);
            for (i = 0; i < len; i++)
                zway_log(zway, Debug, ZSTR("  [%02d] %f"), (int)i, float_arr[i]);
            break;
        case ArrayOfString:
            zdata_get_string_array(data, &str_arr, &len);
            zway_log(zway, Debug, ZSTR("DATA %s = string[%d]"), path, (int)len);
            for (i = 0; i < len; i++)
                zway_log(zway, Debug, ZSTR("  [%02d] \"%s\""), (int)i, str_arr[i]);
            break;
    }
    free(path);
    
    ZDataIterator child = zdata_first_child(data);
    while (child != NULL)
    {
        path = zdata_get_path(child->data);
        zway_log(zway, Debug, ZSTR("CHILD %s"), path);
        free(path);
        child = zdata_next_child(child);
    }
}

void print_basic_holder(const ZWay zway, ZWDataChangeType type, ZDataHolder data)
{
    int int_val;
    
    zdata_get_integer(data, &int_val);
    zway_log(zway, Debug, ZSTR("Basic set value = %i"), int_val); 
}

void print_D_I_CC_event(const ZWay zway, ZWDeviceChangeType type, ZWNODE node_id, ZWBYTE instance_id, ZWBYTE command_id, void *arg)
{
    switch (type)
    {
        case  DeviceAdded:
            zway_log(zway, Information, ZSTR("New device added: %i"), node_id);
            break;
        
        case DeviceRemoved:
            zway_log(zway, Information, ZSTR("Device removed: %i"), node_id);
            break;
        
        case InstanceAdded:
            zway_log(zway, Information, ZSTR("New instance added to device %i: %i"), node_id, instance_id);
            break;
        
        case InstanceRemoved:
            zway_log(zway, Information, ZSTR("Instance removed from device %i: %i"), node_id, instance_id);
            break;
        
        case CommandAdded:
            zway_log(zway, Information, ZSTR("New Command Class added to device %i:%i: %i"), node_id, instance_id, command_id);
            break;
        
        case CommandRemoved:        
            zway_log(zway, Information, ZSTR("Command Class removed from device %i:%i: %i"), node_id, instance_id, command_id);
            break;
    }        
}

void print_help(void) {
    printf("=== Test commands ===\n"
           "d r <path>\n"
           "d d <dev> <path>\n"
           "d i <dev> <inst> <path>\n"
           "d c <dev> <inst> <ccId> <path>\n"
           "m Memory test\n"
           "s <dev> <inst> <ccId> (s=set|g=get) <val>\n"
           "a Start AddNodeToNetwork\n"
           "A Stop AddNodeToNetwork\n"
           "e Start RemoveNodeFromNetwork\n"
           "E Stop RemoveNodeFromNetwork\n"
           "2 k Grant all requested S2 keys\n"
           "2 p Accept S2 pin (DSK)\n"
           "S save configuration to tar file\n"
           "R restore configuration from tar file\n"
           "W restore stick from config\n"
           "x exit\n");
}

int do_work(ZWay zway)
{
    print_help();

    char cmd, cc_cmd, holder_root;
    ZWNODE dev;
    ZWBYTE inst, cc, cc_val, nconv;
    char data_path[256];
    char cmd_buffer[256];
    
    ZWBOOL was_idle = FALSE;
    
    ZWBOOL basic_level_attached = FALSE;
    
    int skip = 0;
    int running = TRUE;    
    while (running)
    {
        if (!zway_is_running(zway))
        {
            running = FALSE;
            break;
        }
        
        if (!zway_is_idle(zway)) 
        {
            sleep_ms(10);
            continue;
        }
        skip = 0;

        if (!basic_level_attached)
        {
            ZDataHolder basic_level_holder;
            zdata_acquire_lock(ZDataRoot(zway));
            basic_level_holder = zway_find_device_instance_cc_data(zway, 8, 0, 0x20, "mylevel");
            if (basic_level_holder)
            {
                basic_level_attached = (zdata_add_callback(basic_level_holder, (ZDataChangeCallback) print_basic_holder, FALSE, NULL) == NoError);
                if (basic_level_attached)
                    zway_log(zway, Debug, ZSTR("Basic.data.mylevel holder handler attached to device 8 instance 0"));
            }
            zdata_release_lock(ZDataRoot(zway));
        }        
                
        printf("> ");
        fgets(cmd_buffer, 255, stdin);
        was_idle = FALSE;        
        
        nconv = sscanf(cmd_buffer, "%c %*s", &cmd);
        if (nconv > 0)
        {
            switch (cmd)
            {
                case 'h':
                    print_help();
                    break;
                
                case 'd':
                    nconv = sscanf(cmd_buffer, "%c %c", &cmd, &holder_root);
                    if (nconv > 1)
                    {
                        switch(holder_root)
                        {
                            case 'r':
                                nconv = sscanf(cmd_buffer, "%c %c %s", &cmd, &holder_root, data_path);
                                if (nconv >= 2)
                                {
                                    if (nconv == 2)
                                    {
                                        data_path[0] = '.';
                                        data_path[1] = '\0';
                                    }
                                    zdata_acquire_lock(ZDataRoot(zway));
                                    dump_data(zway, zway_find_controller_data(zway, data_path));
                                    zdata_release_lock(ZDataRoot(zway));
                                }
                                break;
                                
                            case 'd':
                                nconv = sscanf(cmd_buffer, "%c %c %hd %s", &cmd, &holder_root, &dev, data_path);
                                if (nconv >= 3)
                                {
                                    if (nconv == 3)
                                    {
                                        data_path[0] = '.';
                                        data_path[1] = '\0';
                                    }
                                    zdata_acquire_lock(ZDataRoot(zway));
                                    dump_data(zway, zway_find_device_data(zway, dev, data_path));
                                    zdata_release_lock(ZDataRoot(zway));
                                }
                                break;
                            case 'i':
                                nconv = sscanf(cmd_buffer, "%c %c %hd %hhd %s", &cmd, &holder_root, &dev, &inst, data_path);
                                if (nconv >= 4)
                                {
                                    if (nconv == 4)
                                    {
                                        data_path[0] = '.';
                                        data_path[1] = '\0';
                                    }
                                    zdata_acquire_lock(ZDataRoot(zway));
                                    dump_data(zway, zway_find_device_instance_data(zway, dev, inst, data_path));
                                    zdata_release_lock(ZDataRoot(zway));
                                }
                                break;
                            case 'c':
                                nconv = sscanf(cmd_buffer, "%c %c %hd %hhd %hhd %s", &cmd, &holder_root, &dev, &inst, &cc, data_path);
                                if (nconv >= 5)
                                {
                                    if (nconv == 5)
                                    {
                                        data_path[0] = '.';
                                        data_path[1] = '\0';
                                    }
                                    zdata_acquire_lock(ZDataRoot(zway));
                                    dump_data(zway, zway_find_device_instance_cc_data(zway, dev, inst, cc, data_path));
                                    zdata_release_lock(ZDataRoot(zway));
                                }
                                break;
                        }
                    }
                    break;

                case 's':
                    nconv = sscanf(cmd_buffer, "%c %hd %hhd %hhd %c %hhd", &cmd, &dev, &inst, &cc, &cc_cmd, &cc_val);
                    
                    if (nconv == 6 && cmd == 's' && cc == 0x20 && cc_cmd == 's')
                    {
                        zdata_acquire_lock(ZDataRoot(zway));
                        zway_cc_basic_set(zway, dev, inst, cc_val, NULL, NULL, NULL);
                        zdata_release_lock(ZDataRoot(zway));
                    }
                    else if (nconv == 5 && cmd == 's' && cc == 0x20 && cc_cmd == 'g')
                    {
                        zdata_acquire_lock(ZDataRoot(zway));
                        zway_cc_basic_get(zway, dev, inst, NULL, NULL, NULL);
                        zdata_release_lock(ZDataRoot(zway));
                    }
                    break;
                
                case 'n':
                    nconv = sscanf(cmd_buffer, "%c %hd", &cmd, &dev);
                    
                    if (nconv == 2)
                        zway_fc_request_node_information(zway, dev, NULL, NULL, NULL);
                    break;
                
                case 'N':
                    zway_fc_get_network_stats(zway, NULL, NULL, NULL);
                    break;
                
                case 'm':
                    test_memory(zway);
                    break;

                case 'a':
                    zway_fc_add_node_to_network(zway, TRUE, TRUE, NULL, NULL, NULL);
                    break;

                case '2':
                    nconv = sscanf(cmd_buffer, "%c %hd", &cmd, &dev);
                    
                    if (nconv == 2)
                    {
                        switch(dev) // management functions
                        {
                            case 'k':
                            {
                                int lastIncludedDevice;
                                ZWBOOL res;
                                
                                zdata_acquire_lock(ZDataRoot(zway));
                                (void)zdata_get_integer(zway_find_controller_data(zway, "lastIncludedDevice"), &lastIncludedDevice);
                                if (lastIncludedDevice != 0)
                                {
                                    zdata_get_boolean(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "requestedKeys.S0"), &res);
                                    if (res)
                                        zdata_set_boolean(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "grantedKeys.S0"), TRUE);

                                    zdata_get_boolean(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "requestedKeys.S2Unauthenticated"), &res);
                                    if (res)
                                        zdata_set_boolean(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "grantedKeys.S2Unauthenticated"), TRUE);

                                    zdata_get_boolean(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "requestedKeys.S2Authenticated"), &res);
                                    if (res)
                                        zdata_set_boolean(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "grantedKeys.S2Authenticated"), TRUE);

                                    zdata_get_boolean(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "requestedKeys.S2Access"), &res);
                                    if (res)
                                        zdata_set_boolean(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "grantedKeys.S2Access"), TRUE);

                                    zdata_set_boolean(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "grantedKeys"), TRUE);
                                }
                                zdata_release_lock(ZDataRoot(zway));
                            }
                            break;

                            case 'p':
                            {
                                int lastIncludedDevice;
                                
                                zdata_acquire_lock(ZDataRoot(zway));
                                (void)zdata_get_integer(zway_find_controller_data(zway, "lastIncludedDevice"), &lastIncludedDevice);
                                if (lastIncludedDevice != 0)
                                {
                                    const ZWBYTE *key;
                                    size_t len;
                                    zdata_get_binary(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "publicKey"), &key, &len);
                                    zdata_set_binary(zway_find_device_instance_cc_data(zway, lastIncludedDevice, 0, 0x9f, "publicKeyVerified"), key, len, TRUE);
                                }
                                zdata_release_lock(ZDataRoot(zway));
                            }
                            break;
                        }
                    }
                    break;
                    
                case 'e':
                    zway_fc_remove_node_from_network(zway, TRUE, TRUE, NULL, NULL, NULL);
                    break;
                    
                case 'x':
                    running = FALSE;
                    break;

                case 'S':
                    test_save(zway);
                    break;
                    
                case 'R':
                    test_restore(zway);
                    break;
                    
                case 'l':
                    nconv = sscanf(cmd_buffer, "%c %hd %hhd", &cmd, &dev, &inst);
                    switch(nconv)
                    {
                        case 1:
                            {
                                ZWDevicesList list = zway_devices_list(zway);
                                if (list != NULL) {
                                    int i = 0;
                                    printf("Devices list: ");
                                    while (list[i]) {
                                        printf("%i ", list[i]);
                                        i++;
                                    }
                                    zway_devices_list_free(list);
                                    printf("\n");
                                } else
                                    printf("Error happened requesting devices list\n");
                            }
                            break;

                        case 2:
                            {
                                ZWInstancesList list = zway_instances_list(zway, dev);
                                if (list != NULL) {
                                    int i = 0;
                                    printf("Instances list for device %i: ", dev);
                                    while (list[i]) {
                                        printf("%i ", list[i]);
                                        i++;
                                    }
                                    zway_instances_list_free(list);
                                    printf("\n");
                                } else
                                    printf("Error happened requesting instances list\n");
                            }
                            break;

                        case 3:
                            {
                                ZWCommandClassesList list = zway_command_classes_list(zway, dev, inst);
                                if (list != NULL) {
                                    int i = 0;
                                    printf("Command Classes list for device %i instance %i: ", dev, inst);
                                    while (list[i]) {
                                        printf("%02x ", list[i]);
                                        i++;
                                    }
                                    zway_command_classes_list_free(list);
                                    printf("\n");
                                } else
                                    printf("Error happened requesting command classes list\n");
                            }
                            break;
                    }
                    break;
            }
        }
    }

    return 0;
}

void print_zway_terminated(ZWay zway, void* arg)
{
    zway_log(zway, Information, ZSTR("Z-Way terminated")); 
}


int main (int argc, const char * argv[])
{
    ZWLog logger = zlog_create(stdout, Debug);

    ZWay zway = NULL;
    #ifdef _WINDOWS
    ZWError r = zway_init(&zway, ZSTR("COM3"), 0, NULL, NULL, NULL, NULL, logger);
    #endif
    #ifdef __MACH__
    ZWError r = zway_init(&zway, ZSTR("/dev/cu.SLAB_USBtoUART"), 0, NULL, NULL, NULL, NULL, logger);
    #endif
    #ifdef __FreeBSD__
    ZWError r = zway_init(&zway, ZSTR("/dev/cuaU0"), 0, NULL, NULL, NULL, NULL, logger);
    #endif
    #ifdef __linux__
    ZWError r = zway_init(&zway, ZSTR("/dev/ttyACM0"), 0, NULL, NULL, NULL, NULL, logger);
    #endif
    if (r != NoError)
    {
        zway_log_error(zway, Critical, "Failed to init ZWay", r);
        zlog_close(logger);
        zway_terminate(&zway);
        return -1;
    }

    zway_device_add_callback(zway, DeviceAdded | DeviceRemoved | InstanceAdded | InstanceRemoved | CommandAdded | CommandRemoved, print_D_I_CC_event, NULL);
    
    r = zway_start(zway, print_zway_terminated, NULL);
    if (r != NoError)
    {
        zway_log_error(zway, Critical, "Failed to start ZWay", r);
        zway_terminate(&zway);
        zlog_close(logger);
        return -1;
    }

    r = zway_discover(zway);
    if (r != NoError)
    {
        zway_log_error(zway, Critical, "Failed to negotiate with Z-Wave stick", r);
        zway_terminate(&zway);
        zlog_close(logger);
        return -1;
    }

    // Application code
    int code = do_work(zway);

    r = zway_stop(zway);
    
    zway_terminate(&zway);
    
    zlog_close(logger);
    
    return code;
}

